#!/usr/bin/pypy-1.9

import time
import sys

def fileLen(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1

def init(buffer):
   from glob import globs, setGlobals
   from array import array

   # create spatial variables of specified number
   temp = buffer.pop(0)
   Vars = int(temp.split()[0].strip())+2
   TypeId = temp.split()[1].strip()
   
   # initialize constraints
   N = fileLen(globs['calc']+"/"+globs['calc']+".relations")
   setGlobals("size", N)
   Id = 0
   from bitcoding import B_dict
   with open(globs['calc']+"/"+globs['calc']+".identity") as f:
      Id = B_dict[f.readline().strip()]
   if N <= 8:
      ConMatrix = tuple([array('B',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])
   elif N <= 16:
      ConMatrix = tuple([array('H',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])
   else:
      ConMatrix = tuple([array('I',[B_dict['DALL'] if i != j else Id for i in xrange(Vars)]) for j in xrange(Vars)])
  
   # parse spatial CSP
   from parsecsp import parsecsp
   parsecsp(ConMatrix, buffer)

   return TypeId, ConMatrix

class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg

# main function
def main(argv=None):

   if argv is None:
        argv = sys.argv

   # set some sys useful stuff
   sys.setrecursionlimit(1<<30)
   sys.setcheckinterval(10000)

   from optparse import OptionParser

   # initialize parser for command line arguments
   parser = OptionParser()
   
   # set parsing options
   parser.add_option("-c", "--calculus", type="string", dest="calc", default="rcc8", help="interaction mode: rcc8, or allen [default: %default]")
   parser.add_option("-m", "--method", type="string", dest="method", default="iterative", help="interaction mode: recursive, or iterative [default: %default]")
   parser.add_option("-p", "--pcheuristic", type="string", dest="pcheuristic", default="none", help="interaction mode: none, or weighted [default: %default]")
   parser.add_option("-s", "--split", type="string", dest="split", default="horn", help="interaction mode: base, or horn [default: %default]")
   parser.add_option("-d", "--print", action="store_true", dest="printsol", help="print solution to output")
   parser.add_option("-z", "--pc", action="store_true", dest="pc", help="path consistency only")
   parser.add_option("-t", "--trivial", action="store_true", dest="trivial", help="extra path consistency on the initial graph")
   parser.add_option("-e", "--heuristic", type="string", dest="heuristic", default="global", help="interaction mode: global, local, or none [default: %default]")
   parser.add_option("-f", "--file", action="store", type="string", dest="filename")

   # parse command line arguments
   options, args = parser.parse_args()

   from glob import setGlobals
   #import argparse

   try:   
      # check and set globals for spliting information
      if options.split == "base":
         setGlobals('split',0)
      elif options.split == "horn":
         setGlobals('split',1)
      else:
         raise Usage("Invalid option: " + options.split)

      # check and set globals for calculus information
      if not (options.calc == "rcc8" or options.calc == "allen"):
         raise Usage("Invalid option: " + options.calc)
      else:
         setGlobals('calc',options.calc)

      # check and set globals for patch consistency heuristic information
      if options.pcheuristic == "none":
         setGlobals('pcheuristic',0)
      elif options.pcheuristic == "weighted":
         setGlobals('pcheuristic',1)
      else:
         raise Usage("Invalid option: " + options.pcheuristic)

      # check and set globals for consistency heuristic information
      if options.heuristic == "none":
         setGlobals('heuristic',0)
      elif options.heuristic == "local":
         setGlobals('heuristic',1)
      elif options.heuristic == "global":
         setGlobals('heuristic',2)
      else:
         raise Usage("Invalid option: " + options.heuristic)

      # check and set globals for method decision information
      if options.method == "recursive":
         solver = "rec"    
      elif options.method == "iterative": 
         solver = "iter"
      else:
         raise Usage("Invalid option: " + options.method)

   except Usage, err:
        print err.msg
        parser.print_help()
        return 1

   with open(options.filename) as f:
      total = 0
      meter = 0
      buffer = []

      for i in f:
         buffer.append(i)

         if i.strip() == '.':
            start = elapsed = 0
            meter += 1

            from counters import counter

            TypeId, ConMatrix = init(buffer)
            buffer = []

            from triangulation import MCS, FIC
            from bitcoding import B_dict
               
            edjes = set([])

            neighbors = tuple([set([]) for i in xrange(len(ConMatrix))])
            from bitcoding import B_dict
            DALL = B_dict['DALL']
            for i in xrange(len(ConMatrix)):
               for j in xrange(i+1,len(ConMatrix)):
                  if ConMatrix[i][j] != DALL:
                     edjes.add((i,j))
                     neighbors[i].add(j)
                     neighbors[j].add(i)
   
            start = time.clock()

            from ppc import pPC as ppc

            solution = None
            if (ppc(ConMatrix, neighbors) if options.trivial else True): #path consistency on the initial graph (partial)
   
               fill = FIC(ConMatrix, neighbors, MCS(ConMatrix, neighbors))
   
               neighbors = tuple([set([]) for i in xrange(len(ConMatrix))])

               from bitcoding import B_dict
               DALL = B_dict['DALL']

               incident = [[] for i in xrange(len(ConMatrix))]
               for i,j in fill:
                  incident[j].append(i)  

               from glob import globs

               if globs['pcheuristic'] == 0: # path consistency
                  from ipc import iPC as pc 
               elif globs['pcheuristic'] == 1: # path consistency with weighted queue
                  from ipcw import iPCw as pc
           
               for m, i in enumerate(incident):
                   
                  fills = []
                
                  for n in i:
                     fills.append((n,m))
                     neighbors[m].add(n) 
                     neighbors[n].add(m)               

                  if not pc(ConMatrix,neighbors,fills): 
                     break

               else:
                  solution = ConMatrix
                  if not options.pc:
                     from consistency import consistency
                     from iconsistency import iconsistency
            
                     if solver == "rec": 
                        solution = consistency(ConMatrix, neighbors)
                     elif solver == "iter":
                        solution = iconsistency(ConMatrix, neighbors)

            elapsed = time.clock() - start
            total += elapsed

            # print values of complexity counters
            print '--------------------------------------'
            print 'Type ID: ', TypeId
            print '--------------------------------------'
            print 'Visited nodes: ', counter['nodes']
            print 'Decisions: ', counter['decisions']
            #print 'Visited arcs: ', counter['arcs']
            #print 'Checked constraints: ', counter['cons']
            print '--------------------------------------'
            print 'Time: ', elapsed, 'sec'
            print '--------------------------------------'

            counter['arcs'] = 0
            counter['cons'] = 0
            counter['nodes'] = 0
            counter['decisions'] = 0
  
            # print solution  
            if solution != None:
               print 'Consistent'
               if options.printsol == True:
                  from printm import printMatrix
                  printMatrix(solution)
            else:
               print 'Inconsistent'
            print '--------------------------------------'

            ConMatrix = None
          
   print '======================================'
   print 'Total time:', total, 'sec'
   print 'Avg time per CSP:', total/(1.0*meter), 'sec'
   print '======================================'

if __name__ == '__main__':
   sys.exit(main())

